<?php

App::uses('AppComponent', 'Controller/Component');
App::uses('CakeText', 'Utility');

class BasketComponent extends AppComponent {

	/**
	 * the users hash
	 */
	public $hash = null;

	/**
	 * copy of the basket
	 */
	public $basketCache = null;

	/**
	 * Basket model - set via Component Settings
	 */
	public $Basket = null;

	/**
	 * Basket Item model - set via Component Settings
	 */
	public $BasketItem = null;

	/**
	 * Basket Total model - set via Component Settings
	 */
	public $BasketTotal = null;

	/**
	 * setup the cookie component
	 *
	 */
	public function initialize(Controller $controller) {
		parent::initialize($controller);

		if (empty($this->_controller->BasketCookie)) {
			$this->_controller->BasketCookie = $this->_controller->loadComponent('Cookie');
		}

		$this->_controller->BasketCookie->name = 'EvBasket';
		$this->_controller->BasketCookie->key = 'wX$.NC7-2bf2E~K($"ng8=CyETJY^A~.Nx6sEbwA';
		$this->_controller->BasketCookie->time = WEEK; // CakePHP constant - basics.php - sets timestamp for 1 week
	}

	/**
	 * setup a hash or retrieve a hash from the cookie
	 *
	 * @return string  	Basket hash identifier
	 */
	public function setupHash() {
		if (! empty($this->hash)) {
			return $this->hash;
		}

		if ($this->_controller->BasketCookie->check('basketHash')) {
			return $this->hash = $this->_controller->BasketCookie->read('basketHash');
		}

		$this->hash = $this->generateHash();
		$this->_controller->BasketCookie->write(
			'basketHash',
			$this->hash
		);

		return $this->hash;
	}

	/**
	 * given the hash, setup or return the existing basket
	 *
	 * @param 	string 	$hash 	Basket Hash
	 * @return 	bool 	$result True if a basket was created, false if one already exists
	 */
	public function setupBasket($hash) {
		if (! $this->Basket->checkExists($hash)) {
			return $this->Basket->createBasket(
				$hash
			);
		}

		return false;
	}

	/**
	 * build the basket item contain list
	 *
	 * @param 	array  	$itemContain 	Value from Configure::read('EvBasket.BasketItemContains')
	 * @param 	array  	$relationships 	Value from Configure::read('EvBasket.BasketItemBelongsTo')
	 * @return 	array
	 */
	public function buildItemContain($itemContain, $relationships) {
		if (empty($itemContain)) {
			$itemContain = array();

			foreach ($relationships as $key => $value) {
				if (is_array($value)) {
					$itemContain[] = $key;
				} else {
					$itemContain[] = $value;
				}
			}
		}

		return $itemContain;
	}

	/**
	 * @codeCoverageIgnore
	 * generate a hash for the basket
	 * extracted into another method for easier unit testing
	 *
	 * @return string
	 */
	public function generateHash() {
		return CakeText::uuid();
	}

	/**
	 * build the basket item key
	 *
	 * @param 	array 			$item 	Array with two elements of model and model_id
	 * @return 	string|bool 	$key 	Item basket key Model.Model_id format / false on fail
	 */
	public function buildItemKey($item) {
		if (empty($item['model']) || empty($item['model_id'])) {
			return false;
		}

		return $item['model'] . '.' . $item['model_id'];
	}

	/**
	 * find item key
	 *
	 * @param 	string|int|array 	$key 	Either Basket row ID / compiled basket key string / Array with two elements of model and model_id
	 * @return  string
	 */
	public function getItemKey($key) {
		// set the key as the itemKey as a fallback
		$itemKey = $key;

		if (is_array($key)) {
			// It's an array, try and build it
			$itemKey = $this->buildItemKey($key);
		} elseif (is_numeric($key) && ! empty($this->basketCache)) {
			// it's and id number - try to find the itemKey
			$keysArray = Hash::combine(
				$this->basketCache,
				'BasketItem.{s}.id',
				array(
					'%s.%s',
					'BasketItem.{s}.model',
					'BasketItem.{s}.model_id'
				)
			);

			if (empty($keysArray[$key])) {
				return false;
			}

			$itemKey = $keysArray[$key];
		}

		return $itemKey;
	}

	/**
	 * given the item key, check to see if it exists already in the basket
	 *
	 * @param 	string 		$key 		Item basket key Model.Model_id format
	 * @param 	array|null 	$basket 	Optional basket to search in - if null will use $this->basket
	 * @return 	bool
	 */
	public function basketItemExists($key, $basket = null) {
		if (! empty($basket)) {
			return ! empty($basket['BasketItem'][$key]);
		}

		return ! empty($this->basketCache['BasketItem'][$key]);
	}

	/**
	 * common method to call out to rebuild the totals
	 *
	 * @return bool
	 */
	public function rebuildTotals() {
		// call out to recompile all the basket totals
		$totalResult = $this->BasketTotal->recompileTotals($this->basketCache['Basket']['id']);

		if ($totalResult === true) {
			// dispatch totals recompiled
			$this->_controller->getEventManager()->dispatch(
				new CakeEvent('EvBasket.Component.Basket.postTotalRecompile', $this, array(
					'basketId' => $this->basketCache['Basket']['id']
				))
			);

			// rebuild the basket cache
			$this->getBasket(true);

			return true;
		}

		return false;
	}

	/**
	 * if we're updating an item - check if we actually want to update multi rows in one go
	 *
	 * @param 	array 		$item 		Array with two elements of model and model_id / or multidimensional array of model, model_id, quantity, unitPrice
	 * @param 	int 		$quantity	Amount of items to add
	 * @param 	decimal 	$unitPrice 	Unit price of the item
	 * @return  array 		$toUpdate 	Array of items to update
	 */
	public function buildUpdateList($item, $quantity, $unitPrice) {
		if (is_numeric(key($item))) {
			return $item;
		}

		$item['quantity'] = $quantity;
		$item['unitProce'] = $unitPrice;

		return array(
			$item
		);
	}

	/**
	 * when adding an item, check we have all the data needed
	 *
	 * @param 	array 	$basketItem 	Array of data - array of model, model_id, quantity, unitPrice
	 * @return 	bool
	 */
	public function checkAddItem($item) {
		$Model = EvClassRegistry::init($item['BasketItem']['model']);

		if (
			! empty($item['BasketItem']['model']) &&
			! empty($item['BasketItem']['model_id']) &&
			! empty($item['BasketItem']['quantity']) &&
			method_exists($Model, 'getUnitPrice')
		) {
			return true;
		}

		return false;
	}

	/**
	 * reset a basket - clear the hash and basket
	 *
	 * @return 	bool
	 */
	public function resetBasket() {
		$this->hash = null;
		$this->basketCache = null;

		return true;
	}

	/**
	 * get a single basket item
	 *
	 * @param 	string|int|array 	$key 	Either Basket row ID / compiled basket key string / Array with two elements of model and model_id
	 * @return  array|bool 			$item 	Item array or bool false
	 */
	public function getBasketItem($key) {
		$this->getBasket();

		// set the key as the itemKey as a fallback
		$itemKey = $this->getItemKey($key);

		if ($this->basketItemExists($itemKey)) {
			return $this->basketCache['BasketItem'][$itemKey];
		}

		return false;
	}

	/**
	 * get the basket
	 *
	 * @param 	bool 	$rebuild	Whether to force a rebuild of the basket or not
	 * @return 	array 	$basket 	Basket Data
	 */
	public function getBasket($rebuild = false) {
		$this->setupHash();
		$this->setupBasket($this->hash);

		if ($rebuild === false && ! empty($this->basketCache)) {
			return $this->basketCache;
		}

		$itemContain = $this->buildItemContain(
			Configure::read('EvBasket.BasketItemContains'),
			Configure::read('EvBasket.BasketItemBelongsTo')
		);

		$this->basketCache = $this->Basket->getFullBasket(
			$this->hash,
			$itemContain
		);

		return $this->basketCache;
	}

	/**
	 * add an item to the basket
	 *
	 * @param 	array 		$item 		Array with two elements of model and model_id
	 * @param 	decimal 	$unitPrice 	Unit price of the item
	 * @param 	int 		$quantity	Amount of items to add
	 * @return 	bool
	 */
	public function addItem($item, $unitPrice, $quantity = 1) {
		$this->getBasket();

		$itemKey = $this->buildItemKey($item);

		// check if we are updating an existing item or creating a new row
		if ($this->basketItemExists($itemKey)) {
			$result = $this->BasketItem->updateItem(
				$this->basketCache['BasketItem'][$itemKey],
				($this->basketCache['BasketItem'][$itemKey]['quantity'] + $quantity),
				$unitPrice
			);

		} else {
			$result = $this->BasketItem->addItem(
				$this->basketCache['Basket']['id'],
				$item,
				$unitPrice,
				$quantity
			);
		}

		if ($result === true) {
			// dispatch item added to basket event
			$this->_controller->getEventManager()->dispatch(
				new CakeEvent('EvBasket.Component.Basket.itemAdd', $this, array(
					'basketId' => $this->basketCache['Basket']['id'],
					'itemKey' => $itemKey,
					'BasketItem' => $this->BasketItem
				))
			);

			// call out to recompile all the basket totals
			return $this->rebuildTotals();
		}

		return false;
	}

	/**
	 * update an item in the basket
	 *
	 * @param 	array 		$item 		Array with two elements of model and model_id / or multidimensional array of model, model_id, quantity, unitPrice
	 * @param 	int 		$quantity	Amount of items to add
	 * @param 	decimal 	$unitPrice 	Unit price of the item
	 * @return 	bool
	 */
	public function updateItem($items, $quantity = null, $unitPrice = null) {
		$this->getBasket();

		$toUpdate = $this->buildUpdateList($items, $quantity, $unitPrice);

		$result = false;
		$updated = array();

		//debug($toUpdate);

		foreach ($toUpdate as $item) {
			$itemKey = $this->buildItemKey($item);

			// check if the row actually exists
			if (! $this->basketItemExists($itemKey)) {
				return false;
			}

			// check if we're wanting to delete
			if ($item['quantity'] <= 0) {
				return $this->deleteItem($item);
			}

			//debug($itemKey);

			$updateItem = $this->BasketItem->updateItem(
				$this->basketCache['BasketItem'][$itemKey],
				$item['quantity'],
				(isset($item['unitPrice'])) ? $item['unitPrice'] : null
			);
			//debug($this->BasketItem->validationErrors);
			//debug($updateItem);

			if ($updateItem === true) {
				$result = true;
				$updatedItems[] = $item;
			}
		}

		if ($result === true) {

			// dispatch item added to basket event
			$this->_controller->getEventManager()->dispatch(
				new CakeEvent('EvBasket.Component.Basket.itemUpdated', $this, array(
					'basketId' => $this->basketCache['Basket']['id'],
					'items' => $updatedItems
				))
			);

			// call out to recompile all the basket totals
			return $this->rebuildTotals();
		}

		return false;
	}

	/**
	 * delete and item from the basket
	 *
	 * @param 	array 		$item 		Array with two elements of model and model_id
	 * @return 	bool
	 */
	public function deleteItem($item) {
		$this->getBasket();

		$itemKey = $this->buildItemKey($item);

		// check if the row actually exists
		if (! $this->basketItemExists($itemKey)) {
			return false;
		}

		$result = $this->BasketItem->deleteItem(
			$this->basketCache['BasketItem'][$itemKey]
		);

		if ($result === true) {

			// dispatch item added to basket event
			$this->_controller->getEventManager()->dispatch(
				new CakeEvent('EvBasket.Component.Basket.itemDeleted', $this, array(
					'basketId' => $this->basketCache['Basket']['id'],
					'itemKey' => $itemKey,
					'BasketItem' => $this->BasketItem
				))
			);

			// call out to recompile all the basket totals
			return $this->rebuildTotals();
		}

		return false;
	}
}
